//-----------------------------------------------------------------------------
//
// CO22SnapIoStreamX.cpp
// Copyright (c) 1999 - 2002 by Opto 22
//
// Source for the CO22SnapIoStreamX ActiveX class.  It is mostly a thin 
// wrapper around the O22SnapIoStream C++ class. 
// 
// Some of the functions in this file are NOT members of the CO22SnapIoStreamX
// class, but they do work with it. They are called in the secondary thread.
//-----------------------------------------------------------------------------

#include "stdafx.h"
#include <comdef.h>

#ifdef WIN32
#ifdef _WINDOWS
#include "OptoSnapIoMemMapX.h"
#endif
#endif


#include "CO22SnapIoMemMapX.h"

#include "CO22SnapIoStreamX.h"

#include "CO22SnapIoVariantUtil.h"


/////////////////////////////////////////////////////////////////////////////
// CO22SnapIoStreamX


int StartStreamCallback(void * pUserParam)
//-------------------------------------------------------------------------------------------------
// Called after the listening thread is started in OpenStreaming().
// We use this opportunity to get the COM interface from the main thread so we can fire
// the event.
//
// This function is called by the secondary thread.
//-------------------------------------------------------------------------------------------------
{
  HRESULT hResult;

  // Get the pointer to the main thread
  CO22SnapIoStreamX * pThis = (CO22SnapIoStreamX*) pUserParam;

  if (pThis)
  {
    CoInitialize(NULL);

    // See if we need to get the proxy
    if (pThis->m_pProxy == NULL)
    {
      pThis->m_pProxy.Release();

      // Unmarshall the proxy from the main thread
      hResult = CoGetInterfaceAndReleaseStream(pThis->g_pStm, STREAM_INTERFACE_ID, 
                                               (void**)(&(pThis->m_pProxy)));
    }
  }

  return SIOMM_OK;
}


int StreamEventCallback(int nTCPIPAddress, void * pUserParam, int nResult)
//-------------------------------------------------------------------------------------------------
// Event callback.
//
// This function is called by the secondary thread.
//-------------------------------------------------------------------------------------------------
{
  // Get the pointer to the main thread
  CO22SnapIoStreamX * pThis = (CO22SnapIoStreamX*) pUserParam;

  // Fire the event
  pThis->FireStreamEvent(nTCPIPAddress, nResult);

  return SIOMM_OK;
}


int StopStreamCallback(void * pUserParam)
//-------------------------------------------------------------------------------------------------
// Called after the listening thread is stopped in CloseStreaming().
// We use this opportunity to get the COM interface from the main thread so we can fire
// the event.
// 
// This function is called by the secondary thread.
//-------------------------------------------------------------------------------------------------
{
  // Get the pointer to the main thread
  CO22SnapIoStreamX * pThis = (CO22SnapIoStreamX*) pUserParam;

  // Cleanup
  if (pThis->m_pProxy)
  {   
    pThis->m_pProxy.Release();
  }

  CoUninitialize();

  return SIOMM_OK;
}


int CO22SnapIoStreamX::FireStreamEvent(int nIPAddress, int nResult)
//-------------------------------------------------------------------------------------------------
// Fire the event.  
//
// This function is called by the secondary thread.
//-------------------------------------------------------------------------------------------------
{
  if (m_pProxy)
  {
    m_pProxy.p->_HandleStreamEvent(nIPAddress, nResult);  
  }
  else
  {
    m_pProxy = NULL; // didn't get the proxy!
  }

  return SIOMM_OK;
}


STDMETHODIMP CO22SnapIoStreamX::OpenStreaming(int nType, int nLength, int nPort, int * pnResult)
//-------------------------------------------------------------------------------------------------
// Prepares to start listening for stream packets from SNAP Ethernet I/O units.
//-------------------------------------------------------------------------------------------------
{
  USES_CONVERSION; // for string conversion

  HRESULT    hResult;

  // We need to allow the second thread to fire events. This is done by marshalling the interface
  // from the primary to secondary (listening) thread.
  hResult = CoMarshalInterThreadInterfaceInStream(STREAM_INTERFACE_ID, 
                                                  (STREAM_INTERFACE_NAME*)(void*)this, &g_pStm);

  // Register the different callback functions. These are used to handle the stream events as
  // well as the setup and cleanup of the COM marshalling.
  m_Stream.SetCallbackFunctions(StartStreamCallback, this, 
                                StreamEventCallback, this,
                                StopStreamCallback,  this);

  // Prepares to start listening for stream packets from SNAP Ethernet I/O units.
  *pnResult = m_Stream.OpenStreaming(nType, nLength, nPort);

  return S_OK;
}


STDMETHODIMP CO22SnapIoStreamX::CloseStreaming(int *pnResult)
//-------------------------------------------------------------------------------------------------
// Stops all listening and cleans everything up.
//-------------------------------------------------------------------------------------------------
{

  // Close it up
  if (SIOMM_OK == *pnResult)
    *pnResult = m_Stream.CloseStreaming();

  return S_OK;
}


STDMETHODIMP CO22SnapIoStreamX::_HandleStreamEvent(int nIPAddress, int nResult)
//-------------------------------------------------------------------------------------------------
// This function is called by the secondary thread to fire the event.
//-------------------------------------------------------------------------------------------------
{
  MessageBeep(-1);

  Fire_OnStreamEvent(nIPAddress, nResult); 

  return S_OK;
}


STDMETHODIMP CO22SnapIoStreamX::GetLastStreamStandardBlock(int * pnIpAddressArg,
                                                           VARIANT * pfAnalogValues, 
                                                           VARIANT * pnDigPointFeature, 
                                                           int * pnStatePts63to32, 
                                                           int * pnStatePts31to0, 
                                                           int * pnOnLatchStatePts63to32,   
                                                           int * pnOnLatchStatePts31to0, 
                                                           int * pnOffLatchStatePts63to32,  
                                                           int * pnOffLatchStatePts31to0, 
                                                           int * pnActiveCountersPts63to32, 
                                                           int * pnActiveCountersPts31to0, 
                                                           int * pnResult)
//-------------------------------------------------------------------------------------------------
// A thin wrapper around the C++ function
//-------------------------------------------------------------------------------------------------
{
  // This function is nearly identical to "GetStreamReadArea()"
  CHECK_VARIANT_IS_FLOAT_ARRAY((*pfAnalogValues));
  CHECK_VARIANT_IS_LONG_ARRAY((*pnDigPointFeature));

  // Check that the arrays are big enough
  if ((pfAnalogValues->parray->rgsabound->cElements    < SIOMM_ABANK_MAX_ELEMENTS)  || 
      (pnDigPointFeature->parray->rgsabound->cElements < SIOMM_DBANK_MAX_FEATURE_ELEMENTS))
  {
    return E_INVALIDARG;
  }
  
  SAFEARRAY * pSafeArray;

  SIOMM_StreamStandardBlock StreamData;

  *pnResult = m_Stream.GetLastStreamStandardBlockEx(&StreamData);

  // If no error on the read, go ahead and unpack the structure
  if (SIOMM_OK == *pnResult)
  {
    int i;
    float * pfArray;
    int  * pnArray;

    // Analog values
    pSafeArray = pfAnalogValues->parray;
    SafeArrayAccessData(pSafeArray, (void**)&pfArray);
    for (i = 0 ; i < SIOMM_ABANK_MAX_ELEMENTS ; i++)
    {
      pfArray[i] = StreamData.fAnalogValue[i];
    }
    SafeArrayUnaccessData(pSafeArray);


    // Digital point features
    pSafeArray = pnDigPointFeature->parray;
    SafeArrayAccessData(pSafeArray, (void**)&pnArray);
    for (i = 0 ; i < SIOMM_DBANK_MAX_FEATURE_ELEMENTS ; i++)
    {
      pnArray[i] = StreamData.nDigPointFeature[i];
    }
    SafeArrayUnaccessData(pSafeArray);

    *pnIpAddressArg = StreamData.nTCPIPAddress;
    *pnStatePts63to32 = StreamData.nStatePts63to32; 
    *pnStatePts31to0 = StreamData.nStatePts31to0;
    *pnOnLatchStatePts63to32 = StreamData.nOnLatchStatePts63to32;   
    *pnOnLatchStatePts31to0 = StreamData.nOnLatchStatePts31to0;
    *pnOffLatchStatePts63to32 = StreamData.nOffLatchStatePts63to32;
    *pnOffLatchStatePts31to0 = StreamData.nOffLatchStatePts31to0;
    *pnActiveCountersPts63to32 = StreamData.nActiveCountersPts63to32;
    *pnActiveCountersPts31to0 = StreamData.nActiveCountersPts31to0;
  }

  return S_OK;
}

STDMETHODIMP CO22SnapIoStreamX::StartStreamListening(BSTR bstrIpAddressArg, 
                                                     int  nTimeoutMS, int * pnResult)
//-------------------------------------------------------------------------------------------------
// A thin wrapper around the C++ function
//-------------------------------------------------------------------------------------------------
{
  USES_CONVERSION; // for string conversion

  *pnResult = m_Stream.StartStreamListening(OLE2A(bstrIpAddressArg), nTimeoutMS);

  return S_OK;
}


STDMETHODIMP CO22SnapIoStreamX::StopStreamListening(BSTR bstrIpAddressArg,  int * pnResult)
//-------------------------------------------------------------------------------------------------
// A thin wrapper around the C++ function
//-------------------------------------------------------------------------------------------------
{
  USES_CONVERSION; // for string conversion

  *pnResult = m_Stream.StopStreamListening(OLE2A(bstrIpAddressArg));

  return S_OK;
}



STDMETHODIMP CO22SnapIoStreamX::GetLastStreamStandardBlockEx(O22_SIOMM_StreamStandardBlock *pStreamData, 
                                                             int *pnResult)
//-------------------------------------------------------------------------------------------------
// A thin wrapper around the C++ function
//-------------------------------------------------------------------------------------------------
{
  *pnResult = m_Stream.GetLastStreamStandardBlockEx(pStreamData);

  return S_OK;
}


STDMETHODIMP CO22SnapIoStreamX::GetLastStreamCustomBlockEx(O22_SIOMM_StreamCustomBlock *pStreamData,
                                                           int *pnResult)
//-------------------------------------------------------------------------------------------------
// A thin wrapper around the C++ function
//-------------------------------------------------------------------------------------------------
{
  *pnResult = m_Stream.GetLastStreamCustomBlockEx(pStreamData);

  return S_OK;
}

